#!/usr/bin/perl

use v5.10;
binmode STDOUT, ":utf8";
use strict;
use warnings;

use XML::Parser;
use Date::Calc qw(Delta_Days);
use HTML::Entities;
#use Data::Dumper;

my $version = "1.1";

my %fields = ();
my @MCepoch = (1899, 12, 31);
my @paths;		# xml paths

my $file = shift;

die "Can't find file \"$file\""
  unless -f $file;
    
# Map XML path to MC fields and the MC data type
my @path_to_MC_field = (
    # list types
    [ 'Collection::DVD::CountryOfOrigin',			'Country',			\&add_list_item ],
    [ 'Collection::DVD::Genres::Genre',				'Genre', 			\&add_list_item ],
    [ 'Collection::DVD::Audio::AudioTrack::AudioContent',	'Language',			\&add_list_item ],
    [ 'Collection::DVD::Studios::Studio',			'Studios',			\&add_list_item ],

    # single value
    [ 'Collection::DVD::Format::FormatAspectRatio',		'Aspect Ratio',			\&add_single_item ],
    [ 'Collection::DVD::Format::FormatVideoStandard',		'Video Standard',		\&add_single_item ],
    [ 'Collection::DVD::OriginalTitle',				'Original Title',		\&add_single_item ],
    [ 'Collection::DVD::Rating',			 	'MPAA Rating',			\&add_single_item ],
    [ 'Collection::DVD::RatingSystem',				'MPAA Rating Description',	\&add_single_item ],
    [ 'Collection::DVD::Title',			 		'Name',				\&add_single_item ],
    [ 'Collection::DVD::Overview',			 	'Description',			\&add_single_item ],
    [ 'Collection::DVD::RunningTime',			 	'Duration',			\&add_single_item ],

    # custom user fields
    [ 'Collection::DVD::CollectionNumber',		 	'Collection Number',		\&add_single_item ],
);

my $parser = new XML::Parser(ErrorContext => 2);

$parser->setHandlers(
    Char =>	\&char_handler,
    Start =>	\&start_handler,
    End =>	\&end_handler,
    Final =>	\&final_handler,
    Default =>	\&default_handler);

sub start_handler {
    my $p = shift;
    my $el = shift;

    my $path = join('::', $p->context, $el);
    push @paths, $path;
    if ($path eq 'Collection::DVD') {
	%fields = ();
    }

    sub get_values {
	my $attr_re = shift;
	my $ret = '';

	while (@_) {
	    my ($a, $v) = (shift, shift);
	    $ret = ($ret ? join($1 ? '=' : ' ', $ret, $v) : $v)  if $v and $a =~ /$attr_re/;
	}
	return $ret;
    }

    my $parent = $p->current_element;
    if (defined $parent) {

	if ($path =~ /DVD::Discs::Disc$/) {
	    #say "START DISC: ";
	    push @{$fields{'Filename'}}, '';
	}

	elsif ($path eq 'Collection::DVD::Actors::Actor') {
	    push @{$fields{'Actors'}}, get_values('^(?:First|Middle|Last)Name', @_);
	}

	elsif ($path eq 'Collection::DVD::Credits::Credit') {
	    my $val = get_values('^(?:(?:First|Middle|Last)Name)|(CreditSubtype)$', @_);
	    my $key = ($val =~ s/=(Composer|Director|Screenwriter)$//) ? $1 : 'Credits';
	    add_list_item_unique($key, $val);
	    add_list_item_unique('Artist', $val)	 if $key eq 'Director';	# dual mapped
	    add_list_item_unique('Music By', $val)	 if $key eq 'Composer';	# dual mapped
	}
    }
}

sub add_list_item_unique
{
    push @{$fields{$_[0]}}, $_[1]  if ! exists $fields{$_[0]} or !grep($_[1], split(/;\s*/, @{$fields{$_[0]}}));
}

sub add_list_item
{
    push @{$fields{$_[0]}}, $_[1];
}

sub add_single_item
{
    $fields{$_[0]} .= $_[1];
}

sub char_handler {
    my ($p, $data) = @_;
    my $parent = $p->current_element;

    my $path = $paths[-1];

    # the expat parser returns entities as single characters
    if ($data eq '&' or $data eq '<' or $data eq '>') { $data = encode_entities($data); }
    else					      { $data = decode_entities($data); }

    my ($mcfield,$coderef);
    if (my @found = grep { $path eq $_->[0] } @path_to_MC_field) {
	($mcfield,$coderef) = ($found[0][1], $found[0][2]);

	&$coderef($mcfield, $data);
    }

    elsif ($path =~ /DVD::Audio::AudioTrack::AudioChannels$/) {	# ignoring language for now
	my ($d1,$d2) = ($data =~ /^(\d+)\.(\d)$/);
	$fields{'Channels'} = $d1 + $d2  if $d1;
    }
    elsif ($path =~ /DVD::ProductionYear$/) {
	$fields{'Date'} = 1 + Delta_Days(@MCepoch, ($data,1,1)) if $data ne '0';
    }
    elsif ($path =~ /DVD::ID$/) {
	$fields{'Image File'} = $data . 'f.jpg';
    }

    elsif ($path =~ /DVD::Discs::Disc::Location$/) {
	#say "Filename: ", $data;
	if ($data) {
	    $fields{'Filename'}[-1] .= $data;
	}
    }

    else {
	#print "Parent: $parent, " if defined($parent);
	#say "Data: $data";
    }
}

sub end_handler {
    my $p = shift;
    my $el = shift;

    my $path = pop @paths;
    if ($path eq 'Collection::DVD') {
	$\ = "\r\n";

	$fields{'Media Type'}		= 'Video';
	$fields{'Media Sub Type'}	= 'Movie';

	for (@{$fields{'Filename'}}) {
	    print '<Item>';
	    for my $key (sort keys %fields) {
		print '<Field Name="', $key, '">',
		    ($key eq 'Filename' ? $_ : 
			  ref $fields{$key} eq 'ARRAY' ? join('; ', @{$fields{$key}}) : $fields{$key}),
		'</Field>';
	    }
	    print '</Item>';
	}
    }

    elsif ($path =~ /DVD::Discs::Disc$/) {
	#say "END DISC";
	if (exists $fields{'Filename'}[-1]) {
	    if ($fields{'Filename'}[-1] eq '') {
		# remove non-file disc placeholders
		pop @{$fields{'Filename'}};
	    }
	    else {
		# sometimes filenames are double-quoted in DVD Profiler.
		$fields{'Filename'}[-1] =~ s/"//g
	    }
	}
    }
}

sub final_handler {
    #print Dumper(\%fields);
}

sub default_handler {
}

print '<?xml version="1.0" encoding="UTF-8" standalone="yes" ?>', "\r\n";
print '<MPL Version="2.0" Title="JRSidecar">', "\r\n";

$parser->parsefile($file);

print "</MPL>\r\n";
